/**
 * @file apiexample.c
 * avcodec API use example.
 *
 * Note that this library only handles codecs (mpeg, mpeg4, etc...),
 * not file formats (avi, vob, etc...). See library 'libavformat' for the
 * format handling 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef HAVE_AV_CONFIG_H
#undef HAVE_AV_CONFIG_H
#endif

#include "libavcodec/avcodec.h"
#include "lavc_wrapper.h"

static int lavc_param_vqscale = 2;
static int lavc_param_gop_size = 10000;

static AVSC_CC void config(WrapperContext *wc, LavcOpts *opt)
{
    AVCodecContext *c;

    AVCodec *codec = avcodec_find_encoder(CODEC_ID_MPEG4);
    if(!codec) {
        fprintf(stderr, "\ncodec not found\n");
        exit(1);
    }

    wc->c = c = avcodec_alloc_context();
    wc->picture = avcodec_alloc_frame();
	
    /* put sample parameters */
    c->width = opt->width;
    c->height = opt->height;
    c->frame_rate = 25;
    c->frame_rate_base= 1;
    c->gop_size = lavc_param_gop_size;
    c->flags |= CODEC_FLAG_PASS1;
    c->flags |= CODEC_FLAG_QSCALE;
	c->flags |= opt->v4mv ? CODEC_FLAG_4MV : 0;
	c->pre_me = opt->preme;
	c->dia_size = opt->dia;
	c->pre_dia_size = opt->predia;
	c->me_cmp = opt->cmp;
	c->me_sub_cmp = opt->subcmp;
	
    c->global_quality = 
    wc->picture->quality = (int)(FF_QP2LAMBDA * lavc_param_vqscale + 0.5);
	
    /* open it */
    if (avcodec_open(c, codec) < 0) {
        fprintf(stderr, "could not open codec\n");
        exit(1);
    }
    
    /* alloc image and dummy output buffer */
    wc->outbuf_size = 300000;
    wc->outbuf = (uint8_t*) malloc(wc->outbuf_size);
}

static AVSC_CC void put_image(WrapperContext *wc, PictureYV12 *inpic)
{
	int out_size;
    AVCodecContext *c = wc->c;
    AVFrame *picture = wc->picture;
    
    memcpy(picture->data, inpic->data, sizeof(picture->data));
    memcpy(picture->linesize, inpic->linesize, sizeof(picture->linesize));

    /* encode the image */
    out_size = avcodec_encode_video(c, wc->outbuf, wc->outbuf_size, picture);
    wc->stats = c->stats_out;
}

static LavcFuncs* funcs = NULL;

static AVSC_CC void finish(WrapperContext *wc)
{
    avcodec_close(wc->c);
    free(wc->outbuf);
    free(wc->c);
    free(wc->picture);
	if(funcs)
		free(funcs);
	funcs = NULL;
}

static int initted = 0;

static AVSC_CC void init()
{
	if(initted)
		return;

    /* must be called before using avcodec lib */
    avcodec_init();

    /* register all the codecs (you can also register only the codec
       you wish to have smaller code */
    //avcodec_register_all();
    register_avcodec(&mpeg4_encoder);
	
	initted = 1;
}

LavcFuncs* AVSC_CC sclavc_funcs()
{
	if(funcs)
		return funcs;
	
	funcs = malloc(sizeof(LavcFuncs));
	funcs->init = &init;
	funcs->config = &config;
	funcs->put_image = &put_image;
	funcs->finish = &finish;
	funcs->lavc_build = LIBAVCODEC_BUILD;
	return funcs;
}

