/*
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "internal.h"
#include "info.h"
#include "lavc_wrapper.h"
#include <stdio.h>


#define MYVERSION "0.12"
#define VERSION_PRINTF "# SClavc %s by Loren Merritt, using libavcodec build %d\n", MYVERSION, funcs->lavc_build


static LavcFuncs* funcs = NULL;
void LoadLAVC(IScriptEnvironment* env);


struct FRAMEINFO
{
	unsigned int frame_no;
	PVideoFrame frame;
	unsigned int pitchY, pitchUV;
	unsigned char *frame_ptrY;
	unsigned char *frame_ptrU;
	unsigned char *frame_ptrV;
};

class SClavc : public GenericVideoFilter
{
public:
	SClavc(PClip _child, const char* _logfile, LavcOpts opt, IScriptEnvironment* _env) :
		GenericVideoFilter(_child), env(_env)
	{
		if(!vi.IsYV12())
			ThrowError("SClavc: requires YV12 source");
		if(_logfile)
		{
			logfile = fopen(_logfile, "w");
			if(logfile == NULL)
				ThrowError("SClavc: failed to open logfile");
			fprintf(logfile, VERSION_PRINTF);
		}
		else
			logfile = NULL;

		prev_frame = 0;
		
		opt.width = vi.width;
		opt.height = vi.height;
		funcs->config(&wc, &opt);
	}
	~SClavc()
	{
		//funcs->finish(&wc);
		if(logfile)
			fclose(logfile);
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* _env);
	
private:
	//// Options
	FILE* logfile;

	//// State
	IScriptEnvironment* env;
	int prev_frame;
    WrapperContext wc;

	//// Methods
	void EncodeFrame(int n);
	FRAMEINFO LoadFrame(int n);
	FRAMEINFO FrameInfo(PVideoFrame frm, int n, bool write=0);
	void CopyFrame(PVideoFrame& to, PVideoFrame& from);

	inline void DrawString(PVideoFrame& dst, int x, int y, const char *s);
	void ThrowError(const char* msg);
	template<class T> void ThrowError(const char* msg, T arg);
};

void SClavc::DrawString(PVideoFrame& dst, int x, int y, const char *s)
{
	if (vi.IsYUY2()) ::DrawStringYUY2(dst, x, y, s);
	else ::DrawString(dst, x, y, s);
}

void SClavc::ThrowError(const char* msg)
{
	env->ThrowError(msg);
}

template<class T>
void SClavc::ThrowError(const char* msg, T arg)
{
	// unsafe: I would use snprintf, but MSVC doesn't have it.
	// no, I don't ever free buf
	char* buf = new char[strlen(msg) + 20];
	sprintf(buf, msg, arg);
	env->ThrowError(buf);
}

PVideoFrame __stdcall SClavc::GetFrame(int n, IScriptEnvironment* _env)
{
	env = _env;

	int i;
	for(i = prev_frame+1; i <= n; i++)
		EncodeFrame(i);
	if(logfile)
		fflush(logfile);

	return child->GetFrame(n, env);
}

void SClavc::EncodeFrame(int n)
{
	_ASSERTE(n == prev_frame+1);
	
	FRAMEINFO frm = LoadFrame(n);
	PictureYV12 pic;
	pic.data[0] = frm.frame_ptrY;
	pic.data[1] = frm.frame_ptrU;
	pic.data[2] = frm.frame_ptrV;
	pic.linesize[0] = frm.pitchY;
	pic.linesize[1] = frm.pitchUV;
	pic.linesize[2] = frm.pitchUV;
	
	funcs->put_image(&wc, &pic);

	if(logfile)
		fputs(wc.stats, logfile);
	char buf[200];
	sprintf(buf, "SClavc: %s", wc.stats);
	OutputDebugString(buf);

	prev_frame = n;
}

void LoadLAVC(IScriptEnvironment* env)
{
	if(funcs)
		return;

	HMODULE plugin = LoadLibrary("SceneChange-lavc.dll");
	if(!plugin)
		env->ThrowError("SClavc: can't load SceneChange-lavc.dll\n");
	sclavc_funcs_t* sclavc_get_funcs = (sclavc_funcs_t*) GetProcAddress(plugin, "sclavc_funcs");
	if(!sclavc_get_funcs)
		env->ThrowError("SClavc: can't find sclavc_funcs() in SceneChange-lavc.dll\n");
	funcs = (*sclavc_get_funcs)();

	funcs->init();
}

FRAMEINFO SClavc::LoadFrame(int n)
{
	return FrameInfo(child->GetFrame(n, env), n);
}

FRAMEINFO SClavc::FrameInfo(PVideoFrame frm, int n, bool write)
{
	FRAMEINFO out;
	out.frame_no = n;
	out.frame = frm;
	if(write)
	{
		env->MakeWritable(&frm);
		out.frame_ptrY = (unsigned char *) out.frame->GetWritePtr(PLANAR_Y);
		out.frame_ptrU = (unsigned char *) out.frame->GetWritePtr(PLANAR_U);
		out.frame_ptrV = (unsigned char *) out.frame->GetWritePtr(PLANAR_V);
	}
	else
	{
		out.frame_ptrY = (unsigned char *) out.frame->GetReadPtr(PLANAR_Y);
		out.frame_ptrU = (unsigned char *) out.frame->GetReadPtr(PLANAR_U);
		out.frame_ptrV = (unsigned char *) out.frame->GetReadPtr(PLANAR_V);
	}
	out.pitchY = out.frame->GetPitch(PLANAR_Y);
	out.pitchUV = out.frame->GetPitch(PLANAR_U);
	return out;
}

void SClavc::CopyFrame(PVideoFrame& to, PVideoFrame& from)
{
	env->BitBlt(to->GetWritePtr(PLANAR_Y),
				to->GetPitch(PLANAR_Y),
				from->GetReadPtr(PLANAR_Y),
				from->GetPitch(PLANAR_Y),
				from->GetRowSize(PLANAR_Y),
				from->GetHeight(PLANAR_Y));
	env->BitBlt(to->GetWritePtr(PLANAR_U),
				to->GetPitch(PLANAR_U),
				from->GetReadPtr(PLANAR_U),
				from->GetPitch(PLANAR_U),
				from->GetRowSize(PLANAR_U),
				from->GetHeight(PLANAR_U));
	env->BitBlt(to->GetWritePtr(PLANAR_V),
				to->GetPitch(PLANAR_V),
				from->GetReadPtr(PLANAR_V),
				from->GetPitch(PLANAR_V),
				from->GetRowSize(PLANAR_V),
				from->GetHeight(PLANAR_V));
}

AVSValue __cdecl Create_SClavc(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	char* logfile = NULL;
	LavcOpts opt;
	
	const char* preset = args[2].AsString("fast");
	if(!strcmp(preset, "slow")) {
		opt.predia = -3;
		opt.preme = 2;
		opt.cmp = opt.subcmp = 2;
		opt.v4mv = 1;
	} else if(!strcmp(preset, "fast")) {
		opt.dia = opt.predia = 1;
		opt.preme = 1;
		opt.cmp = opt.subcmp = 0;
		opt.v4mv = 0;
	} else {
		char* buf = new char[strlen(preset) + 30];
		sprintf(buf, "SClavc: unknown preset: '%s'", preset);
		env->ThrowError(buf);
	}
	
	opt.dia = args[3].AsInt(opt.dia);
	opt.predia = args[4].AsInt(opt.predia);
	opt.preme = args[5].AsInt(opt.preme);
	opt.cmp = args[6].AsInt(opt.cmp);
	opt.subcmp = args[7].AsInt(opt.subcmp);
	opt.v4mv = args[8].AsBool(opt.v4mv);
	
	return new SClavc(args[0].AsClip(),
		args[1].AsString(logfile),
		opt, env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	LoadLAVC(env);
    env->AddFunction("SClavc", "c[log]s[preset]s[dia]i[predia]i[preme]i[cmp]i[subcmp]i[v4mv]b", Create_SClavc, 0);
    return 0;
}

